#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

const char *binname = NULL;

void usage(void) {
    fprintf(stderr,
            "fill0\n"
            "fill a section of a file with zero bytes\n"
            "usage: %s -s start [-e end | -l length] file\n"
            "  - to use hex prefix with 0x\n"
            "  - end offset is noninclusive (-s 2 -e 5 is equivalent to -s 2 -l 3)\n"
            "  - if -e or -l is not provided entire end of file will be filled,\n"
            "    but -s is required\n"
            "\n",
            binname);
    exit(EXIT_FAILURE);
}

long readlong(const char *a) {
    char *eptr;
    if (a[0]=='\0') usage();
    long result = strtol(a, &eptr, 0);
    if (eptr[0] != '\0') usage();
    return result;
}

int main(int argc, char **argv) {
    char *filename = NULL;
    long start = -1, end = -1, length = -1;

    binname = argv[0];

    for (int i=1; i < argc; i++) {
        char *option = argv[i];
        char *optarg = (i+1 < argc ? argv[i+1] : NULL);

        if (!strcmp(option, "-s")) {
            if (!optarg) usage();
            if (start != -1) usage();
            start = readlong(optarg);
            if (start < 0) usage();
            i++;
        }
        else if (!strcmp(option, "-e")) {
            if (!optarg) usage();
            if (end != -1) usage();
            end = readlong(optarg);
            if (end < 0) usage();
            i++;
        }
        else if (!strcmp(option, "-l")) {
            if (!optarg) usage();
            if (length != -1) usage();
            length = readlong(optarg);
            if (length < 0) usage();
            i++;
        }
        else {
            if (filename) usage();
            filename = option;
        }
    }
    if (start == -1) usage();
    if (length != -1) {
        if (end != -1) usage();
        end = start + length;
    }
    if (!filename) usage();

    FILE * const file = fopen(filename, "r+b");
    if (!file) {
        int temperrno = errno;
        fprintf(stderr, "error opening %s: ", filename); 
        errno = temperrno;
        perror(NULL);
        exit(EXIT_FAILURE);
    }

    if (fseek(file, 0, SEEK_END) == -1) {
        perror("seek failed");
        exit(EXIT_FAILURE);
    }

    const long file_length = ftell(file);
    if (file_length == -1) {
        perror("ftell failed");
        exit(EXIT_FAILURE);
    }

    if (end == -1) {
        end = file_length;
    }
    if (end < start) usage();
    if (end > file_length || start >= file_length) {
        fprintf(stderr, "request to write past file end (0x%lx)\n",(unsigned long)file_length);
        exit(EXIT_FAILURE);
    }

    if (fseek(file, start, SEEK_SET) == -1) {
        perror("seek failed");
        exit(EXIT_FAILURE);
    }

    for (long offset = start; offset < end; offset++) {
        static const char buf = 0; /* init unneeded due to staticness... */
        if (fwrite(&buf, 1, 1, file) != 1) {
            perror("write failed");
            exit(EXIT_FAILURE);
        }
    }

    if (fclose(file) != 0) {
        perror("write failed");
        exit(EXIT_FAILURE);
    }

    exit(EXIT_SUCCESS);
}
